import urllib

from datetime import timedelta

import pandas as pd

from bs4 import BeautifulSoup

from serenata_toolbox import log
from serenata_toolbox.datasets.helpers import (
    save_to_csv,
    translate_column,
)


class OfficialMissionsDataset:

    URL = (
        'http://www.camara.leg.br/missao-oficial/missao-pesquisa?'
        'deputado=1&'
        'nome-deputado=&'
        'nome-servidor=&'
        'dati={0}&'
        'datf={1}&'
        'nome-evento='
    )

    def fetch(self, start_date, end_date):
        """
        Fetches official missions within the given date range
        """

        records = []
        for two_months_range in self._generate_ranges(start_date, end_date):
            log.debug(two_months_range)
            for record in self._fetch_missions_for_range(two_months_range[0], two_months_range[1]):
                records.append(record)

        df = pd.DataFrame(records, columns=[
            'participant',
            'destination',
            'subject',
            'start',
            'end',
            'canceled',
            'report_status',
            'report_details_link'
        ])

        translate_column(df, 'report_status', {
            'Disponível': 'Available',
            'Pendente': 'Pending',
            'Em análise': 'Analysing',
            'Não se aplica': 'Does not apply'
        })
        translate_column(df, 'canceled', {
            'Não': 'No',
            'Sim': 'Yes'
        })

        return df.drop_duplicates()

    @staticmethod
    def _generate_ranges(start_date, end_date):
        """
        Generate a list of 2 month ranges for the range requested with an
        intersection between months. This is necessary because we can't search
        for ranges longer than 3 months and the period searched has to encompass
        the whole period of the mission.
        """
        range_start = start_date
        while range_start < end_date:
            range_end = range_start + timedelta(days=60)
            yield (
                range_start.strftime("%d/%m/%Y"),
                range_end.strftime("%d/%m/%Y")
            )
            range_start += timedelta(days=30)

    def _fetch_missions_for_range(self, range_start, range_end):
        url = self.URL.format(range_start, range_end)
        data = urllib.request.urlopen(url)
        soup = BeautifulSoup(data, 'html.parser')

        occurences = soup.findAll('tbody', attrs={'class': 'coresAlternadas'})
        if not occurences:
            return

        table = occurences[0]
        for row in table.find_all('tr', recursive=False):
            cells = row.findAll('td', recursive=False)
            start = cells[0].text.strip()
            end = cells[1].text.strip()
            subject = cells[2].text.strip()
            destination = cells[3].text.strip()
            participant = cells[4].find('span').text.strip()
            report_status = cells[4].find('a')
            report_details_link = None
            if report_status is None:
                report_status = cells[4].find_all('td')[1].text.strip()
            else:
                report_path = report_status['href'].strip().replace("\r\n", '').replace("\t", '')
                report_details_link = "http://www.camara.leg.br" + report_path
                report_status = report_status.text.strip()
            canceled = cells[5].text.strip()

            yield (
                participant,
                destination,
                subject,
                start,
                end,
                canceled,
                report_status,
                report_details_link
            )


def fetch_official_missions(data_dir, start_date, end_date):
    """
    :param data_dir: (str) directory in which the output file will be saved
    :param start_date: (datetime) first date of the range to be scraped
    :param end_date: (datetime) last date of the range to be scraped
    """
    official_missions = OfficialMissionsDataset()
    df = official_missions.fetch(start_date, end_date)
    save_to_csv(df, data_dir, "official-missions")

    return df



import os
import urllib
import xml.etree.ElementTree as ET
import socket
import time

import pandas as pd

from serenata_toolbox import log
from serenata_toolbox.datasets.helpers import (
    save_to_csv,
    translate_column,
    xml_extract_datetime,
    xml_extract_text,
)


class PresencesDataset:

    URL = (
        'http://www.camara.leg.br/SitCamaraWS/sessoesreunioes.asmx/ListarPresencasParlamentar'
        '?dataIni={}'
        '&dataFim={}'
        '&numMatriculaParlamentar={}'
    )

    """
    :param sleep_interval: (integer) the amount of seconds to sleep between requests
    """
    def __init__(self, sleep_interval=2):
        self.sleep_interval = sleep_interval

    def fetch(self, deputies, start_date, end_date):
        """
        :param deputies: (pandas.DataFrame) a dataframe with deputies data
        :param date_start: (str) date in the format dd/mm/yyyy
        :param date_end: (str) date in the format dd/mm/yyyy
        """
        log.debug("Fetching data for {} deputies from {} -> {}".format(len(deputies), start_date, end_date))

        records = self._all_presences(deputies, start_date, end_date)

        df = pd.DataFrame(records, columns=(
            'term',
            'congressperson_document',
            'congressperson_name',
            'party',
            'state',
            'date',
            'present_on_day',
            'justification',
            'session',
            'presence'
        ))
        return self._translate(df)

    def _all_presences(self, deputies, start_date, end_date):
        error_count = 0
        for i, deputy in deputies.iterrows():
            log.debug(i, deputy.congressperson_name, deputy.congressperson_document)
            url = self.URL.format(start_date, end_date, deputy.congressperson_document)
            xml = self._try_fetch_xml(10, url)

            if xml is None:
                error_count += 1
            else:
                root = ET.ElementTree(file=xml).getroot()
                for presence in self._parse_deputy_presences(root):
                    yield presence

            time.sleep(self.sleep_interval)

        log.debug("\nErrored fetching", error_count, "deputy presences")

    def _try_fetch_xml(self, attempts, url):
        while attempts > 0:
            try:
                return urllib.request.urlopen(url, data=None, timeout=10)
            except urllib.error.HTTPError as err:
                log.error("HTTP Error", err.code, "when loading URL", url)
                # 500 seems to be the error code for "no data found for the
                # params provided"
                if err.code == 500:
                    log.info("Skipping [HTTP Status 500] {}".format(url))
                    return None
                time.sleep(self.sleep_interval / 2)
                attempts -= 1
                if attempts > 0:
                    log.info("Trying again", attempts)
                else:
                    log.error("FAIL {}".format(url))
            except socket.error as socketerror:
                log.error("Socket error:", socketerror)
                time.sleep(self.sleep_interval * 10)
                attempts -= 1
                if attempts > 0:
                    log.info("Trying again", attempts)
                else:
                    log.error("FAIL {}".format(url))

    @staticmethod
    def _parse_deputy_presences(root):
        term = xml_extract_text(root, 'legislatura')
        congressperson_document = xml_extract_text(root, 'carteiraParlamentar')
        # Please note that this name contains the party and state
        congressperson_name = xml_extract_text(root, 'nomeParlamentar')
        party = xml_extract_text(root, 'siglaPartido')
        state = xml_extract_text(root, 'siglaUF')

        for day in root.findall('.//dia'):
            date = xml_extract_datetime(day, 'data')
            present_on_day = xml_extract_text(day, 'frequencianoDia')
            justification = xml_extract_text(day, 'justificativa')
            for session in day.findall('.//sessao'):
                yield (
                    term,
                    congressperson_document,
                    congressperson_name,
                    party,
                    state,
                    date,
                    present_on_day,
                    justification,
                    xml_extract_text(session, 'descricao'),
                    xml_extract_text(session, 'frequencia')
                )

    @staticmethod
    def _translate(df):
        translate_column(df, 'presence', {
            'Presença': 'Present',
            'Ausência': 'Absent',
        })

        translate_column(df, 'present_on_day', {
            'Presença (~)': 'Present (~)',
            'Presença': 'Present',
            'Ausência': 'Absent',
            'Ausência justificada': 'Justified absence',
        })

        translate_column(df, 'justification', {
            '': '',
            'Atendimento a Obrigação Político-Partidária':
                'Attending to Political-Party Obligation',
            'Ausência Justificada':
                'Justified absence',
            'Decisão da Mesa':
                'Board decision',
            'Licença para Tratamento de Saúde':
                'Health Care Leave',
            'Missão Autorizada':
                'Authorized Mission',
            'Presença Eletrônica Aferida no Painel':
                'Electronic Presence Measured on the Panel'
        })

        return df


def fetch_presences(data_dir, deputies, date_start, date_end):
    """
    :param data_dir: (str) directory in which the output file will be saved
    :param deputies: (pandas.DataFrame) a dataframe with deputies data
    :param date_start: (str) a date in the format dd/mm/yyyy
    :param date_end: (str) a date in the format dd/mm/yyyy
    """
    presences = PresencesDataset()
    df = presences.fetch(deputies, date_start, date_end)
    save_to_csv(df, data_dir, "presences")

    log.info("Presence records:", len(df))
    log.info("Records of deputies present on a session:", len(df[df.presence == 'Present']))
    log.info("Records of deputies absent from a session:", len(df[df.presence == 'Absent']))

    return df



from decouple import config


AMAZON_REGION = config('AMAZON_REGION', default='nyc3')
AMAZON_BUCKET = config('AMAZON_BUCKET', default='serenata-de-amor-data')
AMAZON_ENDPOINT = config('AMAZON_ENDPOINT', default='https://nyc3.digitaloceanspaces.com')



import os

from datetime import datetime

CSV_PARAMS = {
    'compression': 'xz',
    'encoding': 'utf-8',
    'index': False
}


# Utilities for extracting data from XML files

def xml_extract_text(node, xpath):
    """
    :param node: the node to be queried
    :param xpath: the path to fetch the child node that has the wanted text
    """
    text = node.find(xpath).text
    if text is not None:
        text = text.strip()
    return text


def xml_extract_date(node, xpath, date_format='%d/%m/%Y'):
    """
    :param node: the node to be queried
    :param xpath: the path to fetch the child node that has the wanted date
    """
    return datetime.strptime(xml_extract_text(node, xpath), date_format)


def xml_extract_datetime(node, xpath, datetime_format='%d/%m/%Y %H:%M:%S'):
    """
    :param node: the node to be queried
    :param xpath: the path to fetch the child node that has the wanted datetime
    """
    return datetime.strptime(xml_extract_text(node, xpath), datetime_format)


# Utilities for dealing with Dataframes

def translate_column(df, column, translations):
    """
    :param df: (pandas.Dataframe) the dataframe to be translated
    :param column: (str) the column to be translated
    :param translations: (dict) a dictionary of the strings to be categorized and translated
    """
    df[column] = df[column].astype('category')
    translations = [translations[cat]
                    for cat in df[column].cat.categories]
    df[column].cat.rename_categories(translations, inplace=True)


def save_to_csv(df, data_dir, name):
    today = datetime.strftime(datetime.now(), '%Y-%m-%d')
    file_path = os.path.join(data_dir, '{}-{}.xz'.format(today, name))
    df.to_csv(file_path, **CSV_PARAMS)



import os
import urllib
import xml.etree.ElementTree as ET

import pandas as pd

from serenata_toolbox import log
from serenata_toolbox.datasets.helpers import (
    save_to_csv,
    xml_extract_datetime,
    xml_extract_text,
)


class SessionStartTimesDataset:
    URL = (
        'http://www.camara.leg.br/SitCamaraWS/sessoesreunioes.asmx/ListarPresencasDia'
        '?siglaPartido=&siglaUF='
        '&data={0}'
        '&numMatriculaParlamentar={1}'
    )

    def fetch(self, pivot, session_dates):
        """
        :param pivot: (int) a congressperson document to use as a pivot for scraping the data
        :param session_dates: (list) datetime objects to fetch the start times for
        """

        records = self._all_start_times(pivot, session_dates)
        return pd.DataFrame(records, columns=(
            'date',
            'session',
            'started_at'
        ))

    def _all_start_times(self, pivot, session_dates):
        for date in session_dates:
            log.debug(date.strftime("%d/%m/%Y"))
            file = urllib.request.urlopen(self.URL.format(date.strftime("%d/%m/%Y"), pivot))
            tree = ET.ElementTree(file=file)
            for session in tree.getroot().findall('.//sessaoDia'):
                yield (
                    date,
                    xml_extract_text(session, 'descricao'),
                    xml_extract_datetime(session, 'inicio')
                )


def fetch_session_start_times(data_dir, pivot, session_dates):
    """
    :param data_dir: (str) directory in which the output file will be saved
    :param pivot: (int) congressperson document to use as a pivot for scraping the data
    :param session_dates: (list) datetime objects to fetch the start times for
    """
    session_start_times = SessionStartTimesDataset()
    df = session_start_times.fetch(pivot, session_dates)
    save_to_csv(df, data_dir, "session-start-times")

    log.info("Dates requested:", len(session_dates))
    found = pd.to_datetime(df['date'], format="%Y-%m-%d %H:%M:%S").dt.date.unique()
    log.info("Dates found:", len(found))
    return df



import asyncio
import os

import aiofiles
import aiohttp
from tqdm import tqdm

from serenata_toolbox import settings


MAX_REQUESTS = 4


class RemoteFileNotFound(Exception):

    def __init__(self, url):
            self.url = url

    def __str__(self):
        return f'HTTP request to {self.url} returned a non 200 status'


class Downloader:

    LATEST = (
        '2016-08-08-current-year.xz',
        '2016-08-08-last-year.xz',
        '2016-08-08-previous-years.xz',
        '2016-09-03-companies.xz',
        '2016-11-11-congressperson-relatives.xz',
        '2016-11-19-current-year.xz',
        '2016-11-19-last-year.xz',
        '2016-11-19-previous-years.xz',
        '2016-11-28-congressperson-civil-names.xz',
        '2016-11-29-yelp-companies.xz',
        '2016-12-02-foursquare-companies.xz',
        '2016-12-15-speeches.xz',
        '2016-12-20-impeded-non-profit-entities.xz',
        '2016-12-21-inident-and-suspended-companies.xz',
        '2016-12-21-national-register-punished-companies.xz',
        '2016-12-22-agreements.xz',
        '2016-12-22-amendments.xz',
        '2017-02-15-receipts-texts.xz',
        '2017-03-20-purchase-suppliers.xz',
        '2017-05-21-companies-no-geolocation.xz',
        '2017-05-22-brazilian-cities.csv',
        '2017-05-23-congressperson-details.xz',
        '2017-05-29-deputies.xz',
        '2017-05-29-presences.xz',
        '2017-05-29-sessions.xz',
        '2017-05-29-speeches.xz',
        '2017-06-17-official-missions.xz',
        '2017-07-20-tse-candidates.xz',
        '2018-01-05-reimbursements.xz',
        '2018-02-05-congresspeople-social-accounts.xz'
    )

    def __init__(self, target, **kwargs):
        self.target = os.path.abspath(target)
        if not all((os.path.exists(self.target), os.path.isdir(self.target))):
            msg = '{} does not exist or is not a directory.'
            raise FileNotFoundError(msg.format(self.target))

        self.timeout = kwargs.get('timeout')
        self.semaphore = asyncio.Semaphore(MAX_REQUESTS)
        self.progress = 0
        self.total = 0

    def download(self, files):
        if isinstance(files, str):
            files = [files]

        files = tuple(filter(bool, files))
        if not files:
            return

        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.main(loop, files))

    async def main(self, loop, files):
        desc = 'Downloading {} files'.format(len(files))
        if len(files) == 1:
            first_file, *_ = files
            desc = 'Downloading {}'.format(first_file)

        async with aiohttp.ClientSession(loop=loop) as client:

            # fetch total size (all files)
            sizes = [self.fetch_size(client, f) for f in files]
            await asyncio.gather(*sizes)

            # download
            args = dict(total=self.total, desc=desc, unit='b', unit_scale=True)
            with tqdm(**args) as progress:
                self.progress = progress
                downloads = [self.fetch_file(client, f) for f in files]
                await asyncio.gather(*downloads)

            # cleanup
            del self.progress
            self.total = 0

    async def fetch_size(self, client, filename):
        with (await self.semaphore):
            async with client.head(self.url(filename)) as resp:
                if resp.status != 200:
                    raise RemoteFileNotFound(self.url(filename))

                size = resp.headers.get('CONTENT-LENGTH', '0')

        self.total += int(size)

    async def fetch_file(self, client, filename):
        filepath = os.path.join(self.target, filename)
        with (await self.semaphore):
            async with client.get(self.url(filename), timeout=self.timeout) as resp:
                contents = await resp.read()

            async with aiofiles.open(filepath, 'wb') as fh:
                await fh.write(contents)

            self.progress.update(len(contents))

    def url(self, filename):
        base = f'https://s3-{settings.AMAZON_REGION}.amazonaws.com'
        if settings.AMAZON_ENDPOINT:
            base = settings.AMAZON_ENDPOINT

        return f'{base}/{settings.AMAZON_BUCKET}/{filename}'



import logging
import os
import sys


LEVEL = logging.DEBUG if os.environ.get('DEBUG') else logging.INFO
LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


# get a default logger to the project
log = logging.getLogger()
log.setLevel(LEVEL)

# make this log send messages to stdout
output = logging.StreamHandler(sys.stdout)
output.setLevel(LEVEL)
output.setFormatter(logging.Formatter(LOG_FORMAT))
log.addHandler(output)



from setuptools import setup

REPO_URL = 'http://github.com/okfn-brasil/serenata-toolbox'

with open('README.rst') as fobj:
    long_description = fobj.read()

setup(
    author='Serenata de Amor',
    author_email='contato@serenata.ai',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
    ],
    description='Toolbox for Serenata de Amor project',
    zip_safe=False,
    install_requires=[
        'aiofiles',
        'aiohttp',
        'beautifulsoup4>=4.4',
        'lxml>=3.6',
        'pandas>=0.18',
        'python-decouple>=3.1',
        'tqdm'
    ],
    keywords='serenata de amor, data science, brazil, corruption',
    license='MIT',
    long_description=long_description,
    name='serenata-toolbox',
    packages=[
        'serenata_toolbox',
        'serenata_toolbox.federal_senate',
        'serenata_toolbox.chamber_of_deputies',
        'serenata_toolbox.datasets'
    ],
    scripts=['serenata_toolbox/serenata-toolbox'],
    url=REPO_URL,
    python_requires='>=3.6',
    version='15.1.1',
)
